
/*
 * renderer.js v1.8.0
 * (c) 2017 @Johnny Wu
 * Released under the MIT License.
 */

'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var gfx = _interopDefault(require('gfx.js'));
var vmath = require('vmath');
var memop = require('memop');

var enums = {
  // projection
  PROJ_PERSPECTIVE: 0,
  PROJ_ORTHO: 1,

  // lights
  LIGHT_DIRECTIONAL: 0,
  LIGHT_POINT: 1,
  LIGHT_SPOT: 2,

  // shadows
  SHADOW_NONE: 0,
  SHADOW_HARD: 1,
  SHADOW_SOFT: 2,

  // parameter type
  PARAM_INT:             0,
  PARAM_INT2:            1,
  PARAM_INT3:            2,
  PARAM_INT4:            3,
  PARAM_FLOAT:           4,
  PARAM_FLOAT2:          5,
  PARAM_FLOAT3:          6,
  PARAM_FLOAT4:          7,
  PARAM_COLOR3:          8,
  PARAM_COLOR4:          9,
  PARAM_MAT2:           10,
  PARAM_MAT3:           11,
  PARAM_MAT4:           12,
  PARAM_TEXTURE_2D:     13,
  PARAM_TEXTURE_CUBE:   14,

  // clear flags
  CLEAR_COLOR: 1,
  CLEAR_DEPTH: 2,
  CLEAR_STENCIL: 4,
};

var InputAssembler = function InputAssembler(vb, ib, pt) {
  if ( pt === void 0 ) pt = gfx.PT_TRIANGLES;

  this._vertexBuffer = vb;
  this._indexBuffer = ib;
  this._primitiveType = pt;
  this._start = 0;
  this._count = -1;

  // TODO: instancing data
  // this._stream = 0;
};

InputAssembler.prototype.getPrimitiveCount = function getPrimitiveCount () {
  if (this._count !== -1) {
    return this._count;
  }

  if (this._indexBuffer) {
    return this._indexBuffer.count;
  }

  return this._vertexBuffer.count;
};

var Pass = function Pass(name) {
  this._programName = name;

  // cullmode
  this._cullMode = gfx.CULL_BACK;

  // blending
  this._blend = false;
  this._blendEq = gfx.BLEND_FUNC_ADD;
  this._blendAlphaEq = gfx.BLEND_FUNC_ADD;
  this._blendSrc = gfx.BLEND_ONE;
  this._blendDst = gfx.BLEND_ZERO;
  this._blendSrcAlpha = gfx.BLEND_ONE;
  this._blendDstAlpha = gfx.BLEND_ZERO;
  this._blendColor = 0xffffffff;

  // depth
  this._depthTest = false;
  this._depthWrite = false;
  this._depthFunc = gfx.DS_FUNC_LESS,

  // stencil
  this._stencilTest = false;
  // front
  this._stencilFuncFront = gfx.DS_FUNC_ALWAYS;
  this._stencilRefFront = 0;
  this._stencilMaskFront = 0xff;
  this._stencilFailOpFront = gfx.STENCIL_OP_KEEP;
  this._stencilZFailOpFront = gfx.STENCIL_OP_KEEP;
  this._stencilZPassOpFront = gfx.STENCIL_OP_KEEP;
  this._stencilWriteMaskFront = 0xff;
  // back
  this._stencilFuncBack = gfx.DS_FUNC_ALWAYS;
  this._stencilRefBack = 0;
  this._stencilMaskBack = 0xff;
  this._stencilFailOpBack = gfx.STENCIL_OP_KEEP;
  this._stencilZFailOpBack = gfx.STENCIL_OP_KEEP;
  this._stencilZPassOpBack = gfx.STENCIL_OP_KEEP;
  this._stencilWriteMaskBack = 0xff;
};

Pass.prototype.setCullMode = function setCullMode (cullMode) {
  this._cullMode = cullMode;
};

Pass.prototype.setBlend = function setBlend (
  blendEq,
  blendSrc,
  blendDst,
  blendAlphaEq,
  blendSrcAlpha,
  blendDstAlpha,
  blendColor
) {
    if ( blendEq === void 0 ) blendEq = gfx.BLEND_FUNC_ADD;
    if ( blendSrc === void 0 ) blendSrc = gfx.BLEND_ONE;
    if ( blendDst === void 0 ) blendDst = gfx.BLEND_ZERO;
    if ( blendAlphaEq === void 0 ) blendAlphaEq = gfx.BLEND_FUNC_ADD;
    if ( blendSrcAlpha === void 0 ) blendSrcAlpha = gfx.BLEND_ONE;
    if ( blendDstAlpha === void 0 ) blendDstAlpha = gfx.BLEND_ZERO;
    if ( blendColor === void 0 ) blendColor = 0xffffffff;

  this._blend = true;
  this._blendEq = blendEq;
  this._blendSrc = blendSrc;
  this._blendDst = blendDst;
  this._blendAlphaEq = blendAlphaEq;
  this._blendSrcAlpha = blendSrcAlpha;
  this._blendDstAlpha = blendDstAlpha;
  this._blendColor = blendColor;
};

Pass.prototype.setDepth = function setDepth (
  depthTest,
  depthWrite,
  depthFunc
) {
    if ( depthTest === void 0 ) depthTest = false;
    if ( depthWrite === void 0 ) depthWrite = false;
    if ( depthFunc === void 0 ) depthFunc = gfx.DS_FUNC_LESS;

  this._depthTest = depthTest;
  this._depthWrite = depthWrite;
  this._depthFunc = depthFunc;
};

Pass.prototype.setStencilFront = function setStencilFront (
  stencilFunc,
  stencilRef,
  stencilMask,
  stencilFailOp,
  stencilZFailOp,
  stencilZPassOp,
  stencilWriteMask
) {
    if ( stencilFunc === void 0 ) stencilFunc = gfx.DS_FUNC_ALWAYS;
    if ( stencilRef === void 0 ) stencilRef = 0;
    if ( stencilMask === void 0 ) stencilMask = 0xff;
    if ( stencilFailOp === void 0 ) stencilFailOp = gfx.STENCIL_OP_KEEP;
    if ( stencilZFailOp === void 0 ) stencilZFailOp = gfx.STENCIL_OP_KEEP;
    if ( stencilZPassOp === void 0 ) stencilZPassOp = gfx.STENCIL_OP_KEEP;
    if ( stencilWriteMask === void 0 ) stencilWriteMask = 0xff;

  this._stencilTest = true;
  this._stencilFuncFront = stencilFunc;
  this._stencilRefFront = stencilRef;
  this._stencilMaskFront = stencilMask;
  this._stencilFailOpFront = stencilFailOp;
  this._stencilZFailOpFront = stencilZFailOp;
  this._stencilZPassOpFront = stencilZPassOp;
  this._stencilWriteMaskFront = stencilWriteMask;
};

Pass.prototype.setStencilBack = function setStencilBack (
  stencilFunc,
  stencilRef,
  stencilMask,
  stencilFailOp,
  stencilZFailOp,
  stencilZPassOp,
  stencilWriteMask
) {
    if ( stencilFunc === void 0 ) stencilFunc = gfx.DS_FUNC_ALWAYS;
    if ( stencilRef === void 0 ) stencilRef = 0;
    if ( stencilMask === void 0 ) stencilMask = 0xff;
    if ( stencilFailOp === void 0 ) stencilFailOp = gfx.STENCIL_OP_KEEP;
    if ( stencilZFailOp === void 0 ) stencilZFailOp = gfx.STENCIL_OP_KEEP;
    if ( stencilZPassOp === void 0 ) stencilZPassOp = gfx.STENCIL_OP_KEEP;
    if ( stencilWriteMask === void 0 ) stencilWriteMask = 0xff;

  this._stencilTest = true;
  this._stencilFuncBack = stencilFunc;
  this._stencilRefBack = stencilRef;
  this._stencilMaskBack = stencilMask;
  this._stencilFailOpBack = stencilFailOp;
  this._stencilZFailOpBack = stencilZFailOp;
  this._stencilZPassOpBack = stencilZPassOp;
  this._stencilWriteMaskBack = stencilWriteMask;
};

var _stageOffset = 0;
var _name2stageID = {};

var config = {
  addStage: function (name) {
    // already added
    if (_name2stageID[name] !== undefined) {
      return;
    }

    var stageID = 1 << _stageOffset;
    _name2stageID[name] = stageID;

    _stageOffset += 1;
  },

  stageID: function (name) {
    var id = _name2stageID[name];
    if (id === undefined) {
      return -1;
    }
    return id;
  },

  stageIDs: function (nameList) {
    var key = 0;
    for (var i = 0; i < nameList.length; ++i) {
      var id = _name2stageID[nameList[i]];
      if (id !== undefined) {
        key |= id;
      }
    }
    return key;
  }
};

var _genID = 0;

var Technique = function Technique(stages, parameters, passes, layer) {
  if ( layer === void 0 ) layer = 0;

  this._id = _genID++;
  this._stageIDs = config.stageIDs(stages);
  this._parameters = parameters; // {name, type, size, val}
  this._passes = passes;
  this._layer = layer;
  // TODO: this._version = 'webgl' or 'webgl2' // ????
};

var prototypeAccessors = { passes: {},stageIDs: {} };

Technique.prototype.setStages = function setStages (stages) {
  this._stageIDs = config.stageIDs(stages);
};

prototypeAccessors.passes.get = function () {
  return this._passes;
};

prototypeAccessors.stageIDs.get = function () {
  return this._stageIDs;
};

Object.defineProperties( Technique.prototype, prototypeAccessors );

var Effect = function Effect(techniques, properties, defines) {
  if ( properties === void 0 ) properties = {};
  if ( defines === void 0 ) defines = [];

  this._techniques = techniques;
  this._properties = properties;
  this._defines = defines;

  // TODO: check if params is valid for current technique???
};

Effect.prototype.clear = function clear () {
  this._techniques.length = 0;
  this._properties = null;
  this._defines.length = 0;
};

Effect.prototype.getTechnique = function getTechnique (stage) {
    var this$1 = this;

  var stageID = config.stageID(stage);
  for (var i = 0; i < this._techniques.length; ++i) {
    var tech = this$1._techniques[i];
    if (tech.stageIDs & stageID) {
      return tech;
    }
  }

  return null;
};

Effect.prototype.getProperty = function getProperty (name) {
  return this._properties[name];
};

Effect.prototype.setProperty = function setProperty (name, value) {
  // TODO: check if params is valid for current technique???
  this._properties[name] = value;
};

Effect.prototype.getDefine = function getDefine (name) {
    var this$1 = this;

  for (var i = 0; i < this._defines.length; ++i) {
    var def = this$1._defines[i];
    if ( def.name === name ) {
      return def.value;
    }
  }

  console.warn(("Failed to get define " + name + ", define not found."));
  return null;
};

Effect.prototype.define = function define (name, value) {
    var this$1 = this;

  for (var i = 0; i < this._defines.length; ++i) {
    var def = this$1._defines[i];
    if ( def.name === name ) {
      def.value = value;
      return;
    }
  }

  console.warn(("Failed to set define " + name + ", define not found."));
};

Effect.prototype.extractDefines = function extractDefines (out) {
    var this$1 = this;
    if ( out === void 0 ) out = {};

  for (var i = 0; i < this._defines.length; ++i) {
    var def = this$1._defines[i];
    out[def.name] = def.value;
  }

  return out;
};

/**
 * @param {object} json
 */


/**
 * @param {gfx.Device} device
 * @param {Object} data
 */
function createIA(device, data) {
  if (!data.positions) {
    console.error('The data must have positions field');
    return null;
  }

  var verts = [];
  var vcount = data.positions.length / 3;

  for (var i = 0; i < vcount; ++i) {
    verts.push(data.positions[3 * i], data.positions[3 * i + 1], data.positions[3 * i + 2]);

    if (data.normals) {
      verts.push(data.normals[3 * i], data.normals[3 * i + 1], data.normals[3 * i + 2]);
    }

    if (data.uvs) {
      verts.push(data.uvs[2 * i], data.uvs[2 * i + 1]);
    }
  }

  var vfmt = [];
  vfmt.push({ name: gfx.ATTR_POSITION, type: gfx.ATTR_TYPE_FLOAT32, num: 3 });
  if (data.normals) {
    vfmt.push({ name: gfx.ATTR_NORMAL, type: gfx.ATTR_TYPE_FLOAT32, num: 3 });
  }
  if (data.uvs) {
    vfmt.push({ name: gfx.ATTR_UV0, type: gfx.ATTR_TYPE_FLOAT32, num: 2 });
  }

  var vb = new gfx.VertexBuffer(
    device,
    new gfx.VertexFormat(vfmt),
    gfx.USAGE_STATIC,
    new Float32Array(verts),
    vcount
  );

  var ib = null;
  if (data.indices) {
    ib = new gfx.IndexBuffer(
      device,
      gfx.INDEX_FMT_UINT16,
      gfx.USAGE_STATIC,
      new Uint16Array(data.indices),
      data.indices.length
    );
  }

  return new InputAssembler(vb, ib);
}

var _m4_tmp = vmath.mat4.create();
var _genID$1 = 0;

var View = function View() {
  this._id = _genID$1++;

  // viewport
  this._rect = {
    x: 0, y: 0, w: 1, h: 1
  };

  // TODO:
  // this._scissor = {
  // x: 0, y: 0, w: 1, h: 1
  // };

  // clear options
  this._color = vmath.color4.new(0.3, 0.3, 0.3, 1);
  this._depth = 1;
  this._stencil = 1;
  this._clearFlags = enums.CLEAR_COLOR | enums.CLEAR_DEPTH;

  // matrix
  this._matView = vmath.mat4.create();
  this._matProj = vmath.mat4.create();
  this._matViewProj = vmath.mat4.create();
  this._matInvViewProj = vmath.mat4.create();

  // stages & framebuffer
  this._stages = [];
  this._cullingByID = false;
  this._framebuffer = null;

  this._shadowLight = null; // TODO: should not refer light in view.
};

View.prototype.getForward = function getForward (out) {
  return vmath.vec3.set(
    out,
    -this._matView.m02,
    -this._matView.m06,
    -this._matView.m10
  );
};

View.prototype.getPosition = function getPosition (out) {
  vmath.mat4.invert(_m4_tmp, this._matView);
  return vmath.mat4.getTranslation(out, _m4_tmp);
};

var _forward = vmath.vec3.new(0, 0, -1);

var _m4_tmp$1 = vmath.mat4.create();
var _m3_tmp = vmath.mat3.create();
var _transformedLightDirection = vmath.vec3.create();

// compute light viewProjMat for shadow.
function _computeSpotLightViewProjMatrix(light, outView, outProj) {
  // view matrix
  light._node.getWorldRT(outView);
  vmath.mat4.invert(outView, outView);

  // proj matrix
  vmath.mat4.perspective(outProj, light._spotAngle * light._spotAngleScale, 1, light._shadowMinDepth, light._shadowMaxDepth);
}

function _computeDirectionalLightViewProjMatrix(light, outView, outProj) {
  // view matrix
  light._node.getWorldRT(outView);
  vmath.mat4.invert(outView, outView);

  // TODO: should compute directional light frustum based on rendered meshes in scene.
  // proj matrix
  var halfSize = light._shadowFustumSize / 2;
  vmath.mat4.ortho(outProj, -halfSize, halfSize, -halfSize, halfSize, light._shadowMinDepth, light._shadowMaxDepth);
}

function _computePointLightViewProjMatrix(light, outView, outProj) {
  // TODO:
}

var Light = function Light() {
  this._poolID = -1;
  this._node = null;

  this._type = enums.LIGHT_DIRECTIONAL;

  this._color = vmath.color3.new(1, 1, 1);
  this._intensity = 1;

  // used for spot and point light
  this._range = 1;
  // used for spot light, default to 60 degrees
  this._spotAngle = vmath.toRadian(60);
  this._spotExp = 1;
  // cached for uniform
  this._directionUniform = new Float32Array(3);
  this._positionUniform = new Float32Array(3);
  this._colorUniform = new Float32Array([this._color.r * this._intensity, this._color.g * this._intensity, this._color.b * this._intensity]);
  this._spotUniform = new Float32Array([Math.cos(this._spotAngle * 0.5), this._spotExp]);

  // shadow params
  this._shadowType = enums.SHADOW_NONE;
  this._shadowFrameBuffer = null;
  this._shadowMap = null;
  this._shadowMapDirty = false;
  this._shadowDepthBuffer = null;
  this._shadowResolution = 1024;
  this._shadowBias = 0.00005;
  this._shadowDarkness = 1;
  this._shadowMinDepth = 1;
  this._shadowMaxDepth = 1000;
  this._shadowDepthScale = 50; // maybe need to change it if the distance between shadowMaxDepth and shadowMinDepth is small.
  this._frustumEdgeFalloff = 0; // used by directional and spot light.
  this._viewProjMatrix = vmath.mat4.create();
  this._spotAngleScale = 1; // used for spot light.
  this._shadowFustumSize = 80; // used for directional light.
};

var prototypeAccessors$1 = { color: {},intensity: {},type: {},spotAngle: {},spotExp: {},range: {},shadowType: {},shadowMap: {},viewProjMatrix: {},shadowResolution: {},shadowBias: {},shadowDarkness: {},shadowMinDepth: {},shadowMaxDepth: {},shadowDepthScale: {},frustumEdgeFalloff: {} };

Light.prototype.setNode = function setNode (node) {
  this._node = node;
};

Light.prototype.setColor = function setColor (r, g, b) {
  vmath.color3.set(this._color, r, g, b);
  this._colorUniform[0] = r * this._intensity;
  this._colorUniform[1] = g * this._intensity;
  this._colorUniform[2] = b * this._intensity;
};
prototypeAccessors$1.color.get = function () {
  return this._color;
};

Light.prototype.setIntensity = function setIntensity (val) {
  this._intensity = val;
  this._colorUniform[0] = val * this._color.r;
  this._colorUniform[1] = val * this._color.g;
  this._colorUniform[2] = val * this._color.b;
};
prototypeAccessors$1.intensity.get = function () {
  return this._intensity;
};

Light.prototype.setType = function setType (tpe) {
  this._type = tpe;
};
prototypeAccessors$1.type.get = function () {
  return this._type;
};

Light.prototype.setSpotAngle = function setSpotAngle (val) {
  this._spotAngle = val;
  this._spotUniform[0] = Math.cos(this._spotAngle * 0.5);
};
prototypeAccessors$1.spotAngle.get = function () {
  return this._spotAngle;
};

Light.prototype.setSpotExp = function setSpotExp (val) {
  this._spotExp = val;
  this._spotUniform[1] = val;
};
prototypeAccessors$1.spotExp.get = function () {
  return this._spotExp;
};

Light.prototype.setRange = function setRange (tpe) {
  this._range = tpe;
};
prototypeAccessors$1.range.get = function () {
  return this._range;
};

Light.prototype.setShadowType = function setShadowType (type) {
  if (this._shadowType === enums.SHADOW_NONE && type !== enums.SHADOW_NONE) {
    this._shadowMapDirty = true;
  }
  this._shadowType = type;
};
prototypeAccessors$1.shadowType.get = function () {
  return this._shadowType;
};

prototypeAccessors$1.shadowMap.get = function () {
  return this._shadowMap;
};

prototypeAccessors$1.viewProjMatrix.get = function () {
  return this._viewProjMatrix;
};

Light.prototype.setShadowResolution = function setShadowResolution (val) {
  if (this._shadowResolution !== val) {
    this._shadowMapDirty = true;
  }
  this._shadowResolution = val;
};
prototypeAccessors$1.shadowResolution.get = function () {
  return this._shadowResolution;
};

Light.prototype.setShadowBias = function setShadowBias (val) {
  this._shadowBias = val;
};
prototypeAccessors$1.shadowBias.get = function () {
  return this._shadowBias;
};

Light.prototype.setShadowDarkness = function setShadowDarkness (val) {
  this._shadowDarkness = val;
};
prototypeAccessors$1.shadowDarkness.get = function () {
  return this._shadowDarkness;
};

Light.prototype.setShadowMinDepth = function setShadowMinDepth (val) {
  this._shadowMinDepth = val;
};
prototypeAccessors$1.shadowMinDepth.get = function () {
  if (this._type === enums.LIGHT_DIRECTIONAL) {
    return 1.0;
  }
  return this._shadowMinDepth;
};

Light.prototype.setShadowMaxDepth = function setShadowMaxDepth (val) {
  this._shadowMaxDepth = val;
};
prototypeAccessors$1.shadowMaxDepth.get = function () {
  if (this._type === enums.LIGHT_DIRECTIONAL) {
    return 1.0;
  }
  return this._shadowMaxDepth;
};

Light.prototype.setShadowDepthScale = function setShadowDepthScale (val) {
  this._shadowDepthScale = val;
};
prototypeAccessors$1.shadowDepthScale.get = function () {
  return this._shadowDepthScale;
};

Light.prototype.setFrustumEdgeFalloff = function setFrustumEdgeFalloff (val) {
  this._frustumEdgeFalloff = val;
};
prototypeAccessors$1.frustumEdgeFalloff.get = function () {
  return this._frustumEdgeFalloff;
};

Light.prototype.extractView = function extractView (out, stages) {
  // TODO: view should not handle light.
  out._shadowLight = this;

  // rect
  out._rect.x = 0;
  out._rect.y = 0;
  out._rect.w = this._shadowResolution;
  out._rect.h = this._shadowResolution;

  // clear opts
  vmath.color4.set(out._color, 1, 1, 1, 1);
  out._depth = 1;
  out._stencil = 1;
  out._clearFlags = enums.CLEAR_COLOR | enums.CLEAR_DEPTH;

  // stages & framebuffer
  out._stages = stages;
  out._framebuffer = this._shadowFrameBuffer;

  // view projection matrix
  switch(this._type) {
    case enums.LIGHT_SPOT:
      _computeSpotLightViewProjMatrix(this, out._matView, out._matProj);
      break;

    case enums.LIGHT_DIRECTIONAL:
      _computeDirectionalLightViewProjMatrix(this, out._matView, out._matProj);
      break;

    case enums.LIGHT_POINT:
      _computePointLightViewProjMatrix(this, out._matView, out._matProj);
      break;

    default:
      console.warn('shadow of this light type is not supported');
  }

  // view-projection
  vmath.mat4.mul(out._matViewProj, out._matProj, out._matView);
  this._viewProjMatrix = out._matViewProj;
  vmath.mat4.invert(out._matInvViewProj, out._matViewProj);
};

Light.prototype._updateLightPositionAndDirection = function _updateLightPositionAndDirection () {
  this._node.getWorldMatrix(_m4_tmp$1);
  vmath.mat3.fromMat4(_m3_tmp, _m4_tmp$1);
  vmath.vec3.transformMat3(_transformedLightDirection, _forward, _m3_tmp);
  vmath.vec3.array(this._directionUniform, _transformedLightDirection);
  var pos = this._positionUniform;
  pos[0] = _m4_tmp$1.m12;
  pos[1] = _m4_tmp$1.m13;
  pos[2] = _m4_tmp$1.m14;
};

Light.prototype._generateShadowMap = function _generateShadowMap (device) {
  this._shadowMap = new gfx.Texture2D(device, {
    width: this._shadowResolution,
    height: this._shadowResolution,
    format: gfx.TEXTURE_FMT_RGBA8,
    wrapS: gfx.WRAP_CLAMP,
    wrapT: gfx.WRAP_CLAMP,
  });
  this._shadowDepthBuffer = new gfx.RenderBuffer(device,
    gfx.RB_FMT_D16,
    this._shadowResolution,
    this._shadowResolution
  );
  this._shadowFrameBuffer = new gfx.FrameBuffer(device, this._shadowResolution, this._shadowResolution, {
    colors: [this._shadowMap],
    depth: this._shadowDepthBuffer,
  });
};

Light.prototype._destroyShadowMap = function _destroyShadowMap () {
  if (this._shadowMap) {
    this._shadowMap.destroy();
    this._shadowDepthBuffer.destroy();
    this._shadowFrameBuffer.destroy();
    this._shadowMap = null;
    this._shadowDepthBuffer = null;
    this._shadowFrameBuffer = null;
  }
};

Light.prototype.update = function update (device) {
  this._updateLightPositionAndDirection();

  if (this._shadowType === enums.SHADOW_NONE) {
    this._destroyShadowMap();
  } else if (this._shadowMapDirty) {
    this._destroyShadowMap();
    this._generateShadowMap(device);
    this._shadowMapDirty = false;
  }

};

Object.defineProperties( Light.prototype, prototypeAccessors$1 );

var _matView = vmath.mat4.create();
var _matProj = vmath.mat4.create();
var _matViewProj = vmath.mat4.create();
var _matInvViewProj = vmath.mat4.create();
var _tmp_v3 = vmath.vec3.create();

var Camera = function Camera() {
  this._poolID = -1;
  this._node = null;

  //
  this._projection = enums.PROJ_PERSPECTIVE;

  // clear options
  this._color = vmath.color4.new(0.2, 0.3, 0.47, 1);
  this._depth = 1;
  this._stencil = 1;
  this._clearFlags = enums.CLEAR_COLOR | enums.CLEAR_DEPTH;

  // stages & framebuffer
  this._stages = [];
  this._framebuffer = null;

  // projection properties
  this._near = 0.01;
  this._far = 1000.0;
  this._fov = Math.PI/4.0; // vertical fov
  // this._aspect = 16.0/9.0; // DISABLE: use _rect.w/_rect.h
  this._rect = {
    x: 0, y: 0, w: 1, h: 1
  };

  // ortho properties
  this._orthoHeight = 10;
};

// node
Camera.prototype.getNode = function getNode () {
  return this._node;
};
Camera.prototype.setNode = function setNode (node) {
  this._node = node;
};

// type
Camera.prototype.getType = function getType () {
  return this._projection;
};
Camera.prototype.setType = function setType (type) {
  this._projection = type;
};

// orthoHeight
Camera.prototype.getOrthoHeight = function getOrthoHeight () {
  return this._orthoHeight;
};
Camera.prototype.setOrthoHeight = function setOrthoHeight (val) {
  this._orthoHeight = val;
};

// fov
Camera.prototype.getFov = function getFov () {
  return this._fov;
};
Camera.prototype.setFov = function setFov (fov) {
  this._fov = fov;
};

// near
Camera.prototype.getNear = function getNear () {
  return this._near;
};
Camera.prototype.setNear = function setNear (near) {
  this._near = near;
};

// far
Camera.prototype.getFar = function getFar () {
  return this._far;
};
Camera.prototype.setFar = function setFar (far) {
  this._far = far;
};

// color
Camera.prototype.getColor = function getColor (out) {
  return vmath.color4.copy(out, this._color);
};
Camera.prototype.setColor = function setColor (r, g, b, a) {
  vmath.color4.set(this._color, r, g, b, a);
};

// depth
Camera.prototype.getDepth = function getDepth () {
  return this._depth;
};
Camera.prototype.setDepth = function setDepth (depth) {
  this._depth = depth;
};

// stencil
Camera.prototype.getStencil = function getStencil () {
  return this._stencil;
};
Camera.prototype.setStencil = function setStencil (stencil) {
  this._stencil = stencil;
};

// clearFlags
Camera.prototype.getClearFlags = function getClearFlags () {
  return this._clearFlags;
};
Camera.prototype.setClearFlags = function setClearFlags (flags) {
  this._clearFlags = flags;
};

// rect
Camera.prototype.getRect = function getRect (out) {
  out.x = this._rect.x;
  out.y = this._rect.y;
  out.w = this._rect.w;
  out.h = this._rect.h;

  return out;
};
/**
 * @param {Number} x - [0,1]
 * @param {Number} y - [0,1]
 * @param {Number} w - [0,1]
 * @param {Number} h - [0,1]
 */
Camera.prototype.setRect = function setRect (x, y, w, h) {
  this._rect.x = x;
  this._rect.y = y;
  this._rect.w = w;
  this._rect.h = h;
};

// stages
Camera.prototype.getStages = function getStages () {
  return this._stages;
};
Camera.prototype.setStages = function setStages (stages) {
  this._stages = stages;
};

// framebuffer
Camera.prototype.getFramebuffer = function getFramebuffer () {
  return this._framebuffer;
};
Camera.prototype.setFramebuffer = function setFramebuffer (framebuffer) {
  this._framebuffer = framebuffer;
};

Camera.prototype.extractView = function extractView (out, width, height) {
  // rect
  out._rect.x = this._rect.x * width;
  out._rect.y = this._rect.y * height;
  out._rect.w = this._rect.w * width;
  out._rect.h = this._rect.h * height;

  // clear opts
  out._color = this._color;
  out._depth = this._depth;
  out._stencil = this._stencil;
  out._clearFlags = this._clearFlags;

  // stages & framebuffer
  out._stages = this._stages;
  out._framebuffer = this._framebuffer;

  // view matrix
  this._node.getWorldRT(out._matView);
  vmath.mat4.invert(out._matView, out._matView);

  // projection matrix
  // TODO: if this._projDirty
  var aspect = width / height;
  if (this._projection === enums.PROJ_PERSPECTIVE) {
    vmath.mat4.perspective(out._matProj,
      this._fov,
      aspect,
      this._near,
      this._far
    );
  } else {
    var x = this._orthoHeight * aspect;
    var y = this._orthoHeight;
    vmath.mat4.ortho(out._matProj,
      -x, x, -y, y, this._near, this._far
    );
  }

  // view-projection
  vmath.mat4.mul(out._matViewProj, out._matProj, out._matView);
  vmath.mat4.invert(out._matInvViewProj, out._matViewProj);
};

Camera.prototype.screenToWorld = function screenToWorld (out, screenPos, width, height) {
  var aspect = width / height;
  var cx = this._rect.x * width;
  var cy = this._rect.y * height;
  var cw = this._rect.w * width;
  var ch = this._rect.h * height;

  // view matrix
  this._node.getWorldRT(_matView);
  vmath.mat4.invert(_matView, _matView);

  // projection matrix
  if (this._projection === enums.PROJ_PERSPECTIVE) {
    vmath.mat4.perspective(_matProj,
      this._fov,
      aspect,
      this._near,
      this._far
    );
  } else {
    var x = this._orthoHeight * aspect;
    var y = this._orthoHeight;
    vmath.mat4.ortho(_matProj,
      -x, x, -y, y, this._near, this._far
    );
  }

  // view-projection
  vmath.mat4.mul(_matViewProj, _matProj, _matView);

  // inv view-projection
  vmath.mat4.invert(_matInvViewProj, _matViewProj);

  //
  if (this._projection === enums.PROJ_PERSPECTIVE) {
    // calculate screen pos in far clip plane
    vmath.vec3.set(out,
      (screenPos.x - cx) * 2.0 / cw - 1.0,
      (screenPos.y - cy) * 2.0 / ch - 1.0, // DISABLE: (ch - (screenPos.y - cy)) * 2.0 / ch - 1.0,
      1.0
    );

    // transform to world
    vmath.vec3.transformMat4(out, out, _matInvViewProj);

    //
    this._node.getWorldPos(_tmp_v3);
    vmath.vec3.lerp(out, _tmp_v3, out, screenPos.z / this._far);
  } else {
    var range = this._farClip - this._nearClip;
    vmath.vec3.set(out,
      (screenPos.x - cx) * 2.0 / cw - 1.0,
      (screenPos.y - cy) * 2.0 / ch - 1.0, // DISABLE: (ch - (screenPos.y - cy)) * 2.0 / ch - 1.0,
      (this._far - screenPos.z) / range * 2.0 - 1.0
    );

    // transform to world
    vmath.vec3.transformMat4(out, out, _matInvViewProj);
  }

  return out;
};

Camera.prototype.worldToScreen = function worldToScreen (out, worldPos, width, height) {
  var aspect = width / height;
  var cx = this._rect.x * width;
  var cy = this._rect.y * height;
  var cw = this._rect.w * width;
  var ch = this._rect.h * height;

  // view matrix
  this._node.getWorldRT(_matView);
  vmath.mat4.invert(_matView, _matView);

  // projection matrix
  if (this._projection === enums.PROJ_PERSPECTIVE) {
    vmath.mat4.perspective(_matProj,
      this._fov,
      aspect,
      this._near,
      this._far
    );
  } else {
    var x = this._orthoHeight * aspect;
    var y = this._orthoHeight;
    vmath.mat4.ortho(_matProj,
      -x, x, -y, y, this._near, this._far
    );
  }

  // view-projection
  vmath.mat4.mul(_matViewProj, _matProj, _matView);

  // calculate w
  var w =
    worldPos.x * _matViewProj.m03 +
    worldPos.y * _matViewProj.m07 +
    worldPos.z * _matViewProj.m11 +
    _matViewProj.m15;

  vmath.vec3.transformMat4(out, worldPos, _matViewProj);
  out.x = cx + (out.x / w + 1) * 0.5 * cw;
  out.y = cy + (out.y / w + 1) * 0.5 * ch;

  return out;
};

var Model = function Model() {
  this._poolID = -1;
  this._node = null;
  this._inputAssemblers = [];
  this._effects = [];
  this._defines = [];
  this._dynamicIA = false;
  this._viewID = -1;

  // TODO: we calculate aabb based on vertices
  // this._aabb
};

var prototypeAccessors$2 = { inputAssemblerCount: {},dynamicIA: {},drawItemCount: {} };

prototypeAccessors$2.inputAssemblerCount.get = function () {
  return this._inputAssemblers.length;
};

prototypeAccessors$2.dynamicIA.get = function () {
  return this._dynamicIA;
};

prototypeAccessors$2.drawItemCount.get = function () {
  return this._dynamicIA ? 1 : this._inputAssemblers.length;
};

Model.prototype.setNode = function setNode (node) {
  this._node = node;
};

Model.prototype.setDynamicIA = function setDynamicIA (enabled) {
  this._dynamicIA = enabled;
};

Model.prototype.addInputAssembler = function addInputAssembler (ia) {
  if (this._inputAssemblers.indexOf(ia) !== -1) {
    return;
  }
  this._inputAssemblers.push(ia);
};

Model.prototype.clearInputAssemblers = function clearInputAssemblers () {
  this._inputAssemblers.length = 0;
};

Model.prototype.addEffect = function addEffect (effect) {
  if (this._effects.indexOf(effect) !== -1) {
    return;
  }
  this._effects.push(effect);

  //
  var defs = Object.create(null);
  effect.extractDefines(defs);
  this._defines.push(defs);
};

Model.prototype.clearEffects = function clearEffects () {
  this._effects.length = 0;
  this._defines.length = 0;
};

Model.prototype.extractDrawItem = function extractDrawItem (out, index) {
  if (this._dynamicIA) {
    out.model = this;
    out.node = this._node;
    out.ia = null;
    out.effect = this._effects[0];
    out.defines = out.effect.extractDefines(this._defines[0]);

    return;
  }

  if (index >= this._inputAssemblers.length ) {
    out.model = null;
    out.node = null;
    out.ia = null;
    out.effect = null;
    out.defines = null;

    return;
  }

  out.model = this;
  out.node = this._node;
  out.ia = this._inputAssemblers[index];

  var effect, defines;
  if (index < this._effects.length) {
    effect = this._effects[index];
    defines = this._defines[index];
  } else {
    effect = this._effects[this._effects.length-1];
    defines = this._defines[this._effects.length-1];
  }
  out.effect = effect;
  out.defines = effect.extractDefines(defines);
};

Object.defineProperties( Model.prototype, prototypeAccessors$2 );

var Scene = function Scene() {
  this._lights = new memop.FixedArray(16);
  this._models = new memop.FixedArray(16);
  this._cameras = new memop.FixedArray(16);
  this._debugCamera = null;

  // NOTE: we don't use pool for views (because it's less changed and it doesn't have poolID)
  this._views = [];
};

Scene.prototype._add = function _add (pool, item) {
  if (item._poolID !== -1) {
    return;
  }

  pool.push(item);
  item._poolID = pool.length - 1;
};

Scene.prototype._remove = function _remove (pool, item) {
  if (item._poolID === -1) {
    return;
  }

  pool.data[pool.length-1]._poolID = item._poolID;
  pool.fastRemove(item._poolID);
  item._poolID = -1;
};

Scene.prototype.reset = function reset () {
    var this$1 = this;

  for (var i = 0; i < this._models.length; ++i) {
    var model = this$1._models.data[i];
    model._viewID = -1;
  }
};

Scene.prototype.setDebugCamera = function setDebugCamera (cam) {
  this._debugCamera = cam;
};

// camera

Scene.prototype.getCameraCount = function getCameraCount () {
  return this._cameras.length;
};

Scene.prototype.getCamera = function getCamera (idx) {
  return this._cameras.data[idx];
};

Scene.prototype.addCamera = function addCamera (camera) {
  this._add(this._cameras, camera);
};

Scene.prototype.removeCamera = function removeCamera (camera) {
  this._remove(this._cameras, camera);
};

// model

Scene.prototype.getModelCount = function getModelCount () {
  return this._models.length;
};

Scene.prototype.getModel = function getModel (idx) {
  return this._models.data[idx];
};

Scene.prototype.addModel = function addModel (model) {
  this._add(this._models, model);
};

Scene.prototype.removeModel = function removeModel (model) {
  this._remove(this._models, model);
};

// light

Scene.prototype.getLightCount = function getLightCount () {
  return this._lights.length;
};

Scene.prototype.getLight = function getLight (idx) {
  return this._lights.data[idx];
};

Scene.prototype.addLight = function addLight (light) {
  this._add(this._lights, light);
};

Scene.prototype.removeLight = function removeLight (light) {
  this._remove(this._lights, light);
};

// view

Scene.prototype.addView = function addView (view) {
  if (this._views.indexOf(view) === -1) {
    this._views.push(view);
  }
};

Scene.prototype.removeView = function removeView (view) {
  var idx = this._views.indexOf(view);
  if (idx !== -1) {
    this._views.splice(idx, 1);
  }
};

var _shdID = 0;

function _generateDefines(defs) {
  var defines = [];
  for (var def in defs) {
    if (defs[def] === true) {
      defines.push(("#define " + def));
    }
  }
  return defines.join('\n');
}

function _replaceMacroNums(string, defs) {
  var cache = {};
  var tmp = string;
  for (var def in defs) {
    if (Number.isInteger(defs[def])) {
      cache[def] = defs[def];
    }
  }
  for (var def$1 in cache) {
    var reg = new RegExp(def$1, 'g');
    tmp = tmp.replace(reg, cache[def$1]);
  }
  return tmp;
}

function _unrollLoops(string) {
  var pattern = /#pragma for (\w+) in range\(\s*(\d+)\s*,\s*(\d+)\s*\)([\s\S]+?)#pragma endFor/g;
  function replace(match, index, begin, end, snippet) {
    var unroll = '';
    var parsedBegin = parseInt(begin);
    var parsedEnd = parseInt(end);
    if (parsedBegin.isNaN || parsedEnd.isNaN) {
      console.error('Unroll For Loops Error: begin and end of range must be an int num.');
    }
    for (var i = parsedBegin; i < parsedEnd; ++i) {
      unroll += snippet.replace(new RegExp(("{" + index + "}"), 'g'), i);
    }
    return unroll;
  }
  return string.replace(pattern, replace);
}

var ProgramLib = function ProgramLib(device, templates, chunks) {
  var this$1 = this;
  if ( templates === void 0 ) templates = [];
  if ( chunks === void 0 ) chunks = {};

  this._device = device;
  this._precision = "precision highp float;\n";

  // register templates
  this._templates = {};
  for (var i = 0; i < templates.length; ++i) {
    var tmpl = templates[i];
    this$1.define(tmpl.name, tmpl.vert, tmpl.frag, tmpl.defines);
  }

  // register chunks
  this._chunks = {};
  Object.assign(this._chunks, chunks);

  this._cache = {};
};

/**
 * @param {string} name
 * @param {string} template
 * @param {Array} options
 *
 * @example:
 * programLib.define('foobar', vertTmpl, fragTmpl, [
 *   { name: 'shadow' },
 *   { name: 'lightCount', min: 1, max: 4 }
 * ]);
 */
ProgramLib.prototype.define = function define (name, vert, frag, defines) {
  if (this._templates[name]) {
    console.warn(("Failed to define shader " + name + ": already exists."));
    return;
  }

  var id = ++_shdID;

  // calculate option mask offset
  var offset = 0;
  var loop = function ( i ) {
    var def = defines[i];
    def._offset = offset;

    var cnt = 1;

    if (def.min !== undefined && def.max !== undefined) {
      cnt = Math.ceil((def.max - def.min) * 0.5);

      def._map = function (value) {
        return (value - this._min) << def._offset;
      }.bind(def);
    } else {
      def._map = function (value) {
        if (value) {
          return 1 << def._offset;
        }
        return 0;
      }.bind(def);
    }

    offset += cnt;

    def._offset = offset;
  };

    for (var i = 0; i < defines.length; ++i) loop( i );

  vert = this._precision + vert;
  frag = this._precision + frag;

  // store it
  this._templates[name] = {
    id: id,
    name: name,
    vert: vert,
    frag: frag,
    defines: defines
  };
};

/**
 * @param {string} name
 * @param {Object} options
 */
ProgramLib.prototype.getKey = function getKey (name, defines) {
  var tmpl = this._templates[name];
  var key = 0;
  for (var i = 0; i < tmpl.defines.length; ++i) {
    var tmplDefs = tmpl.defines[i];
    var value = defines[tmplDefs.name];
    if (value === undefined) {
      continue;
    }

    key |= tmplDefs._map(value);
  }

  return key << 8 | tmpl.id;
};

/**
 * @param {string} name
 * @param {Object} options
 */
ProgramLib.prototype.getProgram = function getProgram (name, defines) {
  var key = this.getKey(name, defines);
  var program = this._cache[key];
  if (program) {
    return program;
  }

  // get template
  var tmpl = this._templates[name];
  var customDef = _generateDefines(defines) + '\n';
  var vert = _replaceMacroNums(tmpl.vert, defines);
  vert = customDef + _unrollLoops(vert);
  var frag = _replaceMacroNums(tmpl.frag, defines);
  frag = customDef + _unrollLoops(frag);

  program = new gfx.Program(this._device, {
    vert: vert,
    frag: frag
  });
  program.link();
  this._cache[key] = program;

  return program;
};

var _m3_tmp$1 = vmath.mat3.create();
var _m4_tmp$2 = vmath.mat4.create();

var _stageInfos = new memop.RecyclePool(function () {
  return {
    stage: null,
    items: null,
  };
}, 8);

var _float2_pool = new memop.RecyclePool(function () {
  return new Float32Array(2);
}, 8);

var _float3_pool = new memop.RecyclePool(function () {
  return new Float32Array(3);
}, 8);

var _float4_pool = new memop.RecyclePool(function () {
  return new Float32Array(4);
}, 8);

var _float9_pool = new memop.RecyclePool(function () {
  return new Float32Array(9);
}, 8);

var _float16_pool = new memop.RecyclePool(function () {
  return new Float32Array(16);
}, 8);

var _float64_pool = new memop.RecyclePool(function () {
  return new Float32Array(64);
}, 8);

var _int2_pool = new memop.RecyclePool(function () {
  return new Int32Array(2);
}, 8);

var _int3_pool = new memop.RecyclePool(function () {
  return new Int32Array(3);
}, 8);

var _int4_pool = new memop.RecyclePool(function () {
  return new Int32Array(4);
}, 8);

var _int64_pool = new memop.RecyclePool(function () {
  return new Int32Array(64);
}, 8);

var _type2uniformValue = {};
_type2uniformValue[enums.PARAM_INT] = function (value) {
    return value;
  };
_type2uniformValue[enums.PARAM_INT2] = function (value) {
    return vmath.vec2.array(_int2_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_INT3] = function (value) {
    return vmath.vec3.array(_int3_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_INT4] = function (value) {
    return vmath.vec4.array(_int4_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_FLOAT] = function (value) {
    return value;
  };
_type2uniformValue[enums.PARAM_FLOAT2] = function (value) {
    return vmath.vec2.array(_float2_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_FLOAT3] = function (value) {
    return vmath.vec3.array(_float3_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_FLOAT4] = function (value) {
    return vmath.vec4.array(_float4_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_COLOR3] = function (value) {
    return vmath.color3.array(_float3_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_COLOR4] = function (value) {
    return vmath.color4.array(_float4_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_MAT2] = function (value) {
    return vmath.mat2.array(_float4_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_MAT3] = function (value) {
    return vmath.mat3.array(_float9_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_MAT4] = function (value) {
    return vmath.mat4.array(_float16_pool.add(), value);
  };

var _type2uniformArrayValue = {};
_type2uniformArrayValue[enums.PARAM_INT] = {
    func: function func (values) {
      var result = _int64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        result[i] = values[i];
      }
      return result;
    },
    size: 1,
  };
_type2uniformArrayValue[enums.PARAM_INT2] = {
    func: function func (values) {
      var result = _int64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        result[2 * i] = values[i].x;
        result[2 * i + 1] = values[i].y;
      }
      return result;
    },
    size: 2,
  };
_type2uniformArrayValue[enums.PARAM_INT3] = {
    func: undefined,
    size: 3,
  };
_type2uniformArrayValue[enums.PARAM_INT4] = {
    func: function func (values) {
      var result = _int64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[4 * i] = v.x;
        result[4 * i + 1] = v.y;
        result[4 * i + 2] = v.z;
        result[4 * i + 3] = v.w;
      }
      return result;
    },
    size: 4,
  };
_type2uniformArrayValue[enums.PARAM_FLOAT] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        result[i] = values[i];
      }
      return result;
    },
    size: 1
  };
_type2uniformArrayValue[enums.PARAM_FLOAT2] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        result[2 * i] = values[i].x;
        result[2 * i + 1] = values[i].y;
      }
      return result;
    },
    size: 2,
  };
_type2uniformArrayValue[enums.PARAM_FLOAT3] = {
    func: undefined,
    size: 3,
  };
_type2uniformArrayValue[enums.PARAM_FLOAT4] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[4 * i] = v.x;
        result[4 * i + 1] = v.y;
        result[4 * i + 2] = v.z;
        result[4 * i + 3] = v.w;
      }
      return result;
    },
    size: 4,
  };
_type2uniformArrayValue[enums.PARAM_COLOR3] = {
    func: undefined,
    size: 3,
  };
_type2uniformArrayValue[enums.PARAM_COLOR4] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[4 * i] = v.r;
        result[4 * i + 1] = v.g;
        result[4 * i + 2] = v.b;
        result[4 * i + 3] = v.a;
      }
      return result;
    },
    size: 4,
  };
_type2uniformArrayValue[enums.PARAM_MAT2] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[4 * i] = v.m00;
        result[4 * i + 1] = v.m01;
        result[4 * i + 2] = v.m02;
        result[4 * i + 3] = v.m03;
      }
      return result;
    },
    size: 4
  };
_type2uniformArrayValue[enums.PARAM_MAT3] = {
    func: undefined,
    size: 9
  };
_type2uniformArrayValue[enums.PARAM_MAT4] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[16 * i] = v.m00;
        result[16 * i + 1] = v.m01;
        result[16 * i + 2] = v.m02;
        result[16 * i + 3] = v.m03;
        result[16 * i + 4] = v.m04;
        result[16 * i + 5] = v.m05;
        result[16 * i + 6] = v.m06;
        result[16 * i + 7] = v.m07;
        result[16 * i + 8] = v.m08;
        result[16 * i + 9] = v.m09;
        result[16 * i + 10] = v.m10;
        result[16 * i + 11] = v.m11;
        result[16 * i + 12] = v.m12;
        result[16 * i + 13] = v.m13;
        result[16 * i + 14] = v.m14;
        result[16 * i + 15] = v.m15;
      }
      return result;
    },
    size: 16
  };

var Base = function Base(device, opts) {
  this._device = device;
  this._programLib = new ProgramLib(device, opts.programTemplates, opts.programChunks);
  this._opts = opts;
  this._type2defaultValue = ( obj = {}, obj[enums.PARAM_INT] = 0, obj[enums.PARAM_INT2] = vmath.vec2.new(0, 0), obj[enums.PARAM_INT3] = vmath.vec3.new(0, 0, 0), obj[enums.PARAM_INT4] = vmath.vec4.new(0, 0, 0, 0), obj[enums.PARAM_FLOAT] = 0.0, obj[enums.PARAM_FLOAT2] = vmath.vec2.new(0, 0), obj[enums.PARAM_FLOAT3] = vmath.vec3.new(0, 0, 0), obj[enums.PARAM_FLOAT4] = vmath.vec4.new(0, 0, 0, 0), obj[enums.PARAM_COLOR3] = vmath.color3.new(0, 0, 0), obj[enums.PARAM_COLOR4] = vmath.color4.new(0, 0, 0, 1), obj[enums.PARAM_MAT2] = vmath.mat2.create(), obj[enums.PARAM_MAT3] = vmath.mat3.create(), obj[enums.PARAM_MAT4] = vmath.mat4.create(), obj[enums.PARAM_TEXTURE_2D] = opts.defaultTexture, obj[enums.PARAM_TEXTURE_CUBE] = opts.defaultTextureCube, obj );
  var obj;
  this._stage2fn = {};
  this._usedTextureUnits = 0;

  this._viewPools = new memop.RecyclePool(function () {
    return new View();
  }, 8);

  this._drawItemsPools = new memop.RecyclePool(function () {
    return {
      model: null,
      node: null,
      ia: null,
      effect: null,
      defines: null,
    };
  }, 100);

  this._stageItemsPools = new memop.RecyclePool(function () {
    return new memop.RecyclePool(function () {
      return {
        model: null,
        node: null,
        ia: null,
        effect: null,
        defines: null,
        technique: null,
        sortKey: -1,
      };
    }, 100);
  }, 16);
};

Base.prototype._resetTextuerUnit = function _resetTextuerUnit () {
  this._usedTextureUnits = 0;
};

Base.prototype._allocTextuerUnit = function _allocTextuerUnit () {
  var device = this._device;

  var unit = this._usedTextureUnits;
  if (unit >= device._caps.maxTextureUnits) {
    console.warn(("Trying to use " + unit + " texture units while this GPU supports only " + (device._caps.maxTextureUnits)));
  }

  this._usedTextureUnits += 1;
  return unit;
};

Base.prototype._registerStage = function _registerStage (name, fn) {
  this._stage2fn[name] = fn;
};

Base.prototype._reset = function _reset () {
  this._viewPools.reset();
  this._stageItemsPools.reset();
};

Base.prototype._requestView = function _requestView () {
  return this._viewPools.add();
};

Base.prototype._render = function _render (view, scene) {
    var this$1 = this;

  var device = this._device;

  // setup framebuffer
  device.setFrameBuffer(view._framebuffer);

  // setup viewport
  device.setViewport(
    view._rect.x,
    view._rect.y,
    view._rect.w,
    view._rect.h
  );

  // setup clear
  var clearOpts = {};
  if (view._clearFlags & enums.CLEAR_COLOR) {
    clearOpts.color = [
      view._color.r,
      view._color.g,
      view._color.b,
      view._color.a
    ];
  }
  if (view._clearFlags & enums.CLEAR_DEPTH) {
    clearOpts.depth = view._depth;
  }
  if (view._clearFlags & enums.CLEAR_STENCIL) {
    clearOpts.stencil = view._stencil;
  }
  device.clear(clearOpts);

  // get all draw items
  this._drawItemsPools.reset();

  for (var i = 0; i < scene._models.length; ++i) {
    var model = scene._models.data[i];

    // TODO: HACK: filter model by view
    if (view._cullingByID) {
      if (model._viewID !== view._id) {
        continue;
      }
    } else {
      if (model._viewID !== -1) {
        continue;
      }
    }

    for (var m = 0; m < model.drawItemCount; ++m) {
      var drawItem = this$1._drawItemsPools.add();
      model.extractDrawItem(drawItem, m);
    }
  }

  // TODO: update frustum
  // TODO: visbility test
  // frustum.update(view._viewProj);

  // dispatch draw items to different stage
  _stageInfos.reset();

  for (var i$1 = 0; i$1 < view._stages.length; ++i$1) {
    var stage = view._stages[i$1];
    var stageItems = this$1._stageItemsPools.add();
    stageItems.reset();

    for (var j = 0; j < this._drawItemsPools.length; ++j) {
      var drawItem$1 = this$1._drawItemsPools.data[j];
      var tech = drawItem$1.effect.getTechnique(stage);

      if (tech) {
        var stageItem = stageItems.add();
        stageItem.model = drawItem$1.model;
        stageItem.node = drawItem$1.node;
        stageItem.ia = drawItem$1.ia;
        stageItem.effect = drawItem$1.effect;
        stageItem.defines = drawItem$1.defines;
        stageItem.technique = tech;
        stageItem.sortKey = -1;
      }
    }

    var stageInfo = _stageInfos.add();
    stageInfo.stage = stage;
    stageInfo.items = stageItems;
  }

  // render stages
  for (var i$2 = 0; i$2 < _stageInfos.length; ++i$2) {
    var info = _stageInfos.data[i$2];
    var fn = this$1._stage2fn[info.stage];

    fn(view, info.items);
  }
};

Base.prototype._draw = function _draw (item) {
    var this$1 = this;

  var device = this._device;
  var programLib = this._programLib;
  var node = item.node;
    var ia = item.ia;
    var effect = item.effect;
    var technique = item.technique;
    var defines = item.defines;

  // reset the pool
  // NOTE: we can use drawCounter optimize this
  // TODO: should be configurable
  _float2_pool.reset();
  _float3_pool.reset();
  _float4_pool.reset();
  _float9_pool.reset();
  _float16_pool.reset();
  _float64_pool.reset();
  _int2_pool.reset();
  _int3_pool.reset();
  _int4_pool.reset();
  _int64_pool.reset();

  // set common uniforms
  // TODO: try commit this depends on effect
  // {
  node.getWorldMatrix(_m4_tmp$2);
  device.setUniform('model', vmath.mat4.array(_float16_pool.add(), _m4_tmp$2));

  vmath.mat3.transpose(_m3_tmp$1, vmath.mat3.invert(_m3_tmp$1, vmath.mat3.fromMat4(_m3_tmp$1, _m4_tmp$2)));
  device.setUniform('normalMatrix', vmath.mat3.array(_float9_pool.add(), _m3_tmp$1));
  // }

  // set technique uniforms
  for (var i = 0; i < technique._parameters.length; ++i) {
    var prop = technique._parameters[i];
    var param = effect.getProperty(prop.name);

    if (param === undefined) {
      param = prop.val;
    }

    if (param === undefined) {
      param = this$1._type2defaultValue[prop.type];
    }

    if (param === undefined) {
      console.warn(("Failed to set technique property " + (prop.name) + ", value not found."));
      continue;
    }

    if (
      prop.type === enums.PARAM_TEXTURE_2D ||
      prop.type === enums.PARAM_TEXTURE_CUBE
    ) {
      if (prop.size !== undefined) {
        if (prop.size !== param.length) {
          console.error(("The length of texture array (" + (param.length) + ") is not corrent(expect " + (prop.size) + ")."));
          continue;
        }
        var slots = _int64_pool.add();
        for (var index = 0; index < param.length; ++index) {
          slots[index] = this$1._allocTextuerUnit();
        }
        device.setTextureArray(prop.name, param, slots);
      } else {
        device.setTexture(prop.name, param, this$1._allocTextuerUnit());
      }
    } else {
      var convertedValue = (void 0);
      if (prop.size !== undefined) {
        var convertArray = _type2uniformArrayValue[prop.type];
        if (convertArray.func === undefined) {
          console.error('Uniform array of color3/int3/float3/mat3 can not be supportted!');
          continue;
        }
        if (prop.size * convertArray.size > 64) {
          console.error('Uniform array is too long!');
          continue;
        }
        convertedValue = convertArray.func(param);
      } else {
        var convertFn = _type2uniformValue[prop.type];
        convertedValue = convertFn(param);
      }
      device.setUniform(prop.name, convertedValue);
    }
  }

  // for each pass
  for (var i$1 = 0; i$1 < technique._passes.length; ++i$1) {
    var pass = technique._passes[i$1];
    var count = ia.getPrimitiveCount();

    // set vertex buffer
    device.setVertexBuffer(0, ia._vertexBuffer);

    // set index buffer
    if (ia._indexBuffer) {
      device.setIndexBuffer(ia._indexBuffer);
    }

    // set primitive type
    device.setPrimitiveType(ia._primitiveType);

    // set program
    var program = programLib.getProgram(pass._programName, defines);
    device.setProgram(program);

    // cull mode
    device.setCullMode(pass._cullMode);

    // blend
    if (pass._blend) {
      device.enableBlend();
      device.setBlendFuncSep(
        pass._blendSrc,
        pass._blendDst,
        pass._blendSrcAlpha,
        pass._blendDstAlpha
      );
      device.setBlendEqSep(
        pass._blendEq,
        pass._blendAlphaEq
      );
      device.setBlendColor32(pass._blendColor);
    }

    // depth test & write
    if (pass._depthTest) {
      device.enableDepthTest();
      device.setDepthFunc(pass._depthFunc);
    }
    if (pass._depthWrite) {
      device.enableDepthWrite();
    }

    // stencil
    if (pass._stencilTest) {
      device.enableStencilTest();

      // front
      device.setStencilFuncFront(
        pass._stencilFuncFront,
        pass._stencilRefFront,
        pass._stencilMaskFront
      );
      device.setStencilOpFront(
        pass._stencilFailOpFront,
        pass._stencilZFailOpFront,
        pass._stencilZPassOpFront,
        pass._stencilWriteMaskFront
      );

      // back
      device.setStencilFuncBack(
        pass._stencilFuncBack,
        pass._stencilRefBack,
        pass._stencilMaskBack
      );
      device.setStencilOpBack(
        pass._stencilFailOpBack,
        pass._stencilZFailOpBack,
        pass._stencilZPassOpBack,
        pass._stencilWriteMaskBack
      );
    }

    // draw pass
    device.draw(ia._start, count);

    this$1._resetTextuerUnit();
  }
};

var renderer = {
  // config
  addStage: config.addStage,

  // utils
  createIA: createIA,

  // classes
  Pass: Pass,
  Technique: Technique,
  Effect: Effect,
  InputAssembler: InputAssembler,
  View: View,

  Light: Light,
  Camera: Camera,
  Model: Model,
  Scene: Scene,

  Base: Base,
  ProgramLib: ProgramLib,
};
Object.assign(renderer, enums);

module.exports = renderer;
//# sourceMappingURL=renderer.js.map
